/*
 * =====================================================================================
 *
 *       Filename:  vtk_voting_example.cpp
 *
 *    Description:  test cuda tensor voting library with vtk input
 *
 *        Version:  1.0
 *        Created:  04/18/2012 07:55 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ming Liu (), ming.liu@mavt.ethz.ch
 *        Company:  ETHZ
 *
 * =====================================================================================
 */

#include <iostream>

//#include "CudaVoting.h"
#include "assert.h"
//#include "vtktool.h"
// libpointmatcher
#include "pointmatcher/PointMatcher.h"
#include "pointmatcher/Functions.h"
#include "aliases.h"

#include "tensor_voting.h"

#include <time.h>

using namespace PointMatcherSupport;
using namespace std;
using namespace topomap;

//! A dense matrix over ScalarType
typedef Eigen::Matrix<float, Eigen::Dynamic, Eigen::Dynamic> Matrix;
//! A dense integer matrix
typedef Eigen::Matrix<int, Eigen::Dynamic, Eigen::Dynamic> IntMatrix;

/* 
 * ===  FUNCTION  ============================================================
 *         Name:  Main  
 *  Description:  Entry of program
 * ===========================================================================
 */
int main ( int argc, char *argv[] )
{
    if (argc < 4)
    {
      cout << "Usage: ./fake_voting_sparse_cpu <input.vtk> <output.vtk> sigma" << endl;
      return 0;
    }
    cout << "Input file name: "<< argv[1];
    string filename(argv[1]);
    string outputname(argv[2]);



    DP input;
    input = PM::loadVTK(filename.c_str());
    cout << "Input size: " << input.features.size()/4 << endl;

    // subsampling here. Because the grid will otherwise too slow.
    PM pm;
    PM::DataPointsFilter* randSample;
//    randSample = pm.DataPointsFilterRegistrar.create("RandomSamplingDataPointsFilter",
//                PM::Parameters({
//                    {"prob", "0.05"}                    
//                    }));
//    input = randSample->filter(input);
//    cout << "Sampled size: " << input.features.size()/4 << endl;

    int numPoints = input.features.size()/4;

    // use CPU to vote for planes
    
    float sigma = 0.5;
    if(EOF == sscanf(argv[3], "%f", &sigma))
    {
        cout << "sigma not valid float." << endl;
        return 0;
    }
    cout << "Using sigma size: " << sigma << endl;
    
    TensorVoting voter(sigma); 
    cout << "Set point sets..." << endl;
    voter.setPoints(input);
    cout << "Sparse ball voting (CPU)..." << endl;
    voter.sparse_ball_vote();
    cout	<< "sparse tensor split..." << endl;
    voter.sparse_tensor_split();

    // save sparse tensor
    PM::DataPoints::Descriptors stick=PM::DataPoints::Descriptors::Zero(1, numPoints);
    PM::DataPoints::Descriptors plate=PM::DataPoints::Descriptors::Zero(1, numPoints);
    PM::DataPoints::Descriptors ball =PM::DataPoints::Descriptors::Zero(1, numPoints);
    for(unsigned int i=0; i<numPoints; i++)
    {
        stick(i) = voter.sparseStick(0,i);
        plate(i) = voter.sparsePlate(0,i*2);
        ball(i) =  voter.sparseBall(i);
    }
    input.addDescriptor("stick", stick);
    input.addDescriptor("plate", plate);
    input.addDescriptor("ball", ball);

    PM::saveVTK(input, outputname.c_str());

    return 1;

}

