/*
 * =====================================================================================
 *
 *       Filename:  vtk_voting_example.cpp
 *
 *    Description:  test cuda tensor voting library with vtk input
 *
 *        Version:  1.0
 *        Created:  04/18/2012 07:55 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ming Liu (), ming.liu@mavt.ethz.ch
 *        Company:  ETHZ
 *
 * =====================================================================================
 */

#include <iostream>

//#include "CudaVoting.h"
#include "assert.h"
//#include "vtktool.h"
// libpointmatcher
#include "pointmatcher/PointMatcher.h"
#include "pointmatcher/Functions.h"
#include "aliases.h"

#include "tensor_voting.h"
#include "occupancygrid.h"
#include "vtktool.h"
#include "CudaVoting.h"

#include "boost/tuple/tuple.hpp"
#include <time.h>
#include <cuda.h>
#include <cutil_math.h>
#include <vector_types.h>
#include <iostream>
#include <fstream>
#define USE_GPU_SPARSE_BALL

#define INT_ENDING 10000
//#define INT_ENDING 1000000
//
using namespace PointMatcherSupport;
using namespace std;
using namespace topomap;
using namespace cudavoting;

//! A dense matrix over ScalarType
typedef Eigen::Matrix<float, Eigen::Dynamic, Eigen::Dynamic> Matrix;
//! A dense integer matrix
typedef Eigen::Matrix<int, Eigen::Dynamic, Eigen::Dynamic> IntMatrix;

/* 
 * ===  FUNCTION  ============================================================
 *         Name:  Main  
 *  Description:  Entry of program
 * ===========================================================================
 */
int main ( int argc, char *argv[] )
{
    if (argc < 4)
    {
      cout << "Usage: ./fake_voting_dense_sparse_cpu <input.vtk> <output.vtk> sigma gridsize(1.0)" << endl;
      return 0;
    }
    cout << "Input file name: "<< argv[1];
    string filename(argv[1]);
    string outputname(argv[2]);

    DP input;
    input = PM::loadVTK(filename.c_str());
    cout << "Input size: " << input.features.size()/4 << endl;
    // subsampling here. Because the grid will otherwise too slow.
    PM pm;

    int numPoints = input.features.size()/4;

    // use CPU to vote for planes
    float sigma = 0.5;
    if(EOF == sscanf(argv[3], "%f", &sigma))
    {
        cout << "sigma not valid float." << endl;
        return 0;
    }
    cout << "Using sigma size: " << sigma << endl;



    TensorVoting voter(sigma); 
    cout << "Set point sets..."<< numPoints << endl;
    voter.setPoints(input);
#ifdef USE_GPU_SPARSE_BALL
    cout << "Sparse ball voting (GPU)..." << endl;
    Eigen::Matrix<Eigen::Matrix3f, Eigen::Dynamic, 1> sparseField;
//    CudaVoting::sparseBallVoting(sparseField, input.features, sigma);
//    float3 what = make_float3(1,2,3);

    // set output
    // ATTENTION: sm_11 only support atomicAdd on int, therefore here it need to be int
    // for sm_20 (hardware) or above, the code can be better adapted.
    // for compatiable reasons, here use int

    // float
//    size_t sizeField = numPoints*3*sizeof(float3);
//    float3* h_fieldarray = (float3 *)malloc(sizeField);
//    bzero( h_fieldarray, sizeField );
    // int
    size_t sizeField = numPoints*3*sizeof(int3);
    int3* h_fieldarray = (int3 *)malloc(sizeField);
//    bzero( h_fieldarray, sizeField );

    //set input:   getArrayFromMatrix();
    size_t sizePoints = numPoints*sizeof(float3);
    float3 *h_points = (float3 *)malloc(sizePoints);
    float max_x = -1e10, max_y = -1e10, max_z = -1e10;
    float min_x = 1e10, min_y = 1e10, min_z = 1e10;
    cout << "Filling host point structure... " << endl;
    for(unsigned int i = 0; i<numPoints; i++) // get min_max of coord at the same time
    {
        h_points[i].x = input.features(0,i); 
        h_points[i].y = input.features(1,i); 
        h_points[i].z = input.features(2,i); 
        if (input.features(0,i) > max_x) max_x = input.features(0,i);
        if (input.features(1,i) > max_y) max_y = input.features(1,i);
        if (input.features(2,i) > max_z) max_z = input.features(2,i);
        if (input.features(0,i) < min_x) min_x = input.features(0,i);
        if (input.features(1,i) < min_y) min_y = input.features(1,i);
        if (input.features(2,i) < min_z) min_z = input.features(2,i);
    }

    // set log
    size_t sizeLog = numPoints*sizeof(int2);
    int2 * h_log = (int2 *)malloc(sizeLog);
    bzero( h_log, sizeLog);

    // call CUDA
    cout << "Send to GPU..." << endl;
    CudaVoting::sparseBallVoting(h_fieldarray, h_points, sigma, numPoints, h_log);
    // post-processing
    sparseField.resize(numPoints, 1);
    for(unsigned int i = 0; i<numPoints; i++)
    {
        Eigen::Matrix3f M;
        M << (float)(h_fieldarray[i*3 + 0].x)/INT_ENDING, (float)(h_fieldarray[i*3 + 0].y)/INT_ENDING, (float)(h_fieldarray[i*3 + 0].z)/INT_ENDING, 
             (float)(h_fieldarray[i*3 + 1].x)/INT_ENDING, (float)(h_fieldarray[i*3 + 1].y)/INT_ENDING, (float)(h_fieldarray[i*3 + 1].z)/INT_ENDING, 
             (float)(h_fieldarray[i*3 + 2].x)/INT_ENDING, (float)(h_fieldarray[i*3 + 2].y)/INT_ENDING, (float)(h_fieldarray[i*3 + 2].z)/INT_ENDING;
        sparseField(i) = M;
    }
    voter.setSparseTensor(sparseField);
    printf("Outputing..");

    ofstream out("log_fake_sparse_gpu1.txt");
    for(int i =0; i<numPoints; i++)
        //out << voter.sparseTensors(i) << endl;
        out << sparseField(i) << endl;
    out.close();

    ofstream logf("log_number_of_whowhom_sparse_gpu1.txt");
    for(int i =0; i<numPoints; i++)
    {
        logf << i << " " << h_log[i].x << " " << h_log[i].y << endl;
    }
    logf.close();


    // clean up
    free(h_fieldarray);
    //free(h_points); //???????????????????????????
    free(h_log);

#else
    cout << "Sparse ball voting (CPU)..." << endl;
    voter.sparse_ball_vote();
    cout << endl;
#endif
    cout	<< "sparse tensor split..." << endl;
    voter.sparse_tensor_split();
    cout << endl;

    // save sparse tensor
    PM::DataPoints::Descriptors stick=PM::DataPoints::Descriptors::Zero(1, numPoints);
    PM::DataPoints::Descriptors plate=PM::DataPoints::Descriptors::Zero(1, numPoints);
    PM::DataPoints::Descriptors ball =PM::DataPoints::Descriptors::Zero(1, numPoints);
    for(unsigned int i=0; i<numPoints; i++)
    {
        stick(i) = voter.sparseStick(0,i);
        plate(i) = voter.sparsePlate(0,i*2);
        ball(i) =  voter.sparseBall(i);
    }

#if 1
    // write to pointcloud, then into vtk file
    input.addDescriptor("stick", stick);
    input.addDescriptor("plate", plate);
    input.addDescriptor("ball", ball);
    input.addDescriptor("normals", voter.sparseStick.bottomRows(3));
    PM::saveVTK(input, "sparse_gpu.vtk");
    cout << "Sparse tensor is saved to [sparse_gpu.vtk]" << endl;
#endif



    /*-----------------------------------------------------------------------------
     *  Dense Grid Parameters
     *-----------------------------------------------------------------------------*/
    float cell_size = 0.2;
    if (argc > 4)
      cell_size = atof(argv[4]);
    cout << "Load vtk as grid......cell_size: " << cell_size << endl;

    float hw = 1.0;
    if (argc > 5)
      hw = atof(argv[5]);
    cout << "Using half window " << hw << " size of sigma . Change it with argv[5]." << endl;
    int hw_size = ceil(hw*sigma/cell_size);
    cout << "hw_size is " << hw_size << endl;

    // number of cells in grid
    //size_t maxGridSize = round( 5000.0 / pow(cell_size, 3) );//625000; //0.1*100*50*10/(0.2*0.2*0.2) ;// TODO: modify with point data
    unsigned int grid_dimx = ceil( (max_x - min_x)/ cell_size ) + 2*hw_size +1; // +1 because two ceil function will create 1 additional cell
    unsigned int grid_dimy = ceil( (max_y - min_y)/ cell_size ) + 2*hw_size +1;
    unsigned int grid_dimz = ceil( (max_z - min_z)/ cell_size ) + 2*hw_size +1;
    size_t maxGridSize = grid_dimx * grid_dimy * grid_dimz;
    cout << "@@@@@@@ grid size is " << grid_dimx << " x " << grid_dimy << " x " << grid_dimz << " = " << maxGridSize << "  @@@@@@@@@@" << endl;
    cout << "@@@@@@@ grid ranged from: \n x: [" << min_x << " ," << max_x << "]\n "
        << "y: [" << min_y << " ," << max_y << "]\n "
        << "z: [" << min_z << " ," << max_z << "]\n";

    /*-----------------------------------------------------------------------------
     *  Dense Grid voting
     *-----------------------------------------------------------------------------*/

    // copy to get h_sparsestick as input to GPU
    size_t sizeSparseStick = numPoints*4*sizeof(float4);
    float4 * h_sparsestick = (float4 *)malloc(sizeSparseStick);
    for(unsigned int i = 0; i<numPoints; i++)
    {
        h_sparsestick[i].x = voter.sparseStick(0,i);  // saliency
        h_sparsestick[i].y = voter.sparseStick(1,i);  // x
        h_sparsestick[i].z = voter.sparseStick(2,i);  // y
        h_sparsestick[i].w = voter.sparseStick(3,i);  // z
    }
    // set log
    int2 * hh_log = (int2 *)malloc(sizeLog);
    bzero( hh_log, sizeLog);

    int3 * h_fieldarraydense = (int3 *)malloc(maxGridSize*3*sizeof(int3) ); // around 21MB
    //bzero( h_fieldarraydense, maxGridSize * sizeof(int3) );// setZero of h_field; may not needed.

    // send to GPU
    cout << "Send to GPU dense stick voting... numPoints: "<< numPoints << " gridcells: " << maxGridSize << endl;
    CudaVoting::realDenseGridStickVoting(h_fieldarraydense, h_points, h_sparsestick, 
                maxGridSize, sigma, hw_size, cell_size, numPoints, 
                grid_dimx, grid_dimy, grid_dimz, 
                min_x, min_y, min_z, hh_log); // h_pointgrid is left on GPU only.


    ofstream logff("log_number_of_whowhom_dense_gpu2.txt");
    for(int i =0; i<numPoints; i++)
    {
        logff << i << " " << hh_log[i].x << " " << hh_log[i].y << endl;
    }
    logff.close();

    // post-processing
    cout << "Organize integer conversion... " << endl;
    cout << "Picking non-zero data." << endl;

    vector< Eigen::Matrix3f > denseField; // save dense field of meaningful places
    vector< boost::tuple<float, float, float> > vecDenseCells; // save centers in meter of all meaningful dense cells

    denseField.reserve(maxGridSize);
    vecDenseCells.reserve(maxGridSize);

    // save and grid selection
    for(unsigned int i = 0; i<maxGridSize; i++)
    {
        printf("Finish %% %.2f\r", (float)i*100.0/maxGridSize ); fflush(stdout);
        // check non-zero
        if (h_fieldarraydense[i*3].x == 0 && h_fieldarraydense[i*3].y ==0 && h_fieldarraydense[i*3].z ==0) // first element in matrix is not 0; meaningful tensor
            continue;

        // 1. save dense field
        Eigen::Matrix3f M;
        M.setZero();
        // May need transpose (symmetric anyway)
        M << (float)(h_fieldarraydense[i*3 + 0].x)/INT_ENDING, (float)(h_fieldarraydense[i*3 + 0].y)/INT_ENDING, (float)(h_fieldarraydense[i*3 + 0].z)/INT_ENDING, 
          (float)(h_fieldarraydense[i*3 + 1].x)/INT_ENDING, (float)(h_fieldarraydense[i*3 + 1].y)/INT_ENDING, (float)(h_fieldarraydense[i*3 + 1].z)/INT_ENDING, 
          (float)(h_fieldarraydense[i*3 + 2].x)/INT_ENDING, (float)(h_fieldarraydense[i*3 + 2].y)/INT_ENDING, (float)(h_fieldarraydense[i*3 + 2].z)/INT_ENDING;
        //denseField(i) = M;
        denseField.push_back( M );

        // 2. save cell definition
        // coord refering origin
        int z = i / (grid_dimx*grid_dimy) - hw_size;
        int xy = i % (grid_dimx*grid_dimy);
        int x = xy % grid_dimx - hw_size;
        int y = xy / grid_dimx - hw_size;

        float xx = x*cell_size + 0.5*cell_size + min_x;
        float yy = y*cell_size + 0.5*cell_size + min_y;
        float zz = z*cell_size + 0.5*cell_size + min_z;
        vecDenseCells.push_back( boost::make_tuple(xx, yy, zz) );
    }
    cout << endl;

    cout << "maxGridSize : " << maxGridSize << endl << "realGridSize : " << vecDenseCells.size() << endl;


    /*-----------------------------------------------------------------------------
     *  Take out tensors with non-zero data
     *-----------------------------------------------------------------------------*/
    cout << "Outputing tensor into file: [log_fake_dense_gpu2.txt]" << endl;
    ofstream outdense("log_real_dense_gpu2.txt");
    for(int i =0; i<denseField.size(); i++)
    {
        printf("Finish %% %.2f\r", (float)i*100.0/denseField.size() ); fflush(stdout);
        // save to file
        outdense << denseField[i] << endl;
    }
    cout << endl;
    outdense.close();

    cout	<< "dense tensor split..." << endl;
    // explicit way of tensor decomposition:
    vector<TensorEigenValuesType> tensorVector;
    tensorVector.reserve(denseField.size());
    /*-----------------------------------------------------------------------------
     *  Organize Tensor voting results
     *-----------------------------------------------------------------------------*/
    for(unsigned int i=0; i<denseField.size(); i++)
    {
        printf("Extract eigen values: %d / %d\r", i, denseField.size());
        Eigen::SelfAdjointEigenSolver<Eigen::Matrix3f> es(denseField[i], false); // only calculate eigen values
        Eigen::Vector3f abs_evalues;
        Eigen::Vector3i eig_indices(0,0,0);

        abs_evalues = es.eigenvalues().array().abs();
        abs_evalues.maxCoeff(&eig_indices(0));
        abs_evalues.minCoeff(&eig_indices(2));
        for (unsigned int x=0; x<3; ++x)
        {
            if (x != eig_indices(0) && x!= eig_indices(2)) { eig_indices(1) = x; }
        }	
        TensorEigenValuesType tensorEigenvalues(
                    es.eigenvalues()[eig_indices(0)] - es.eigenvalues()[eig_indices(1)], // stick <- voter for planes
                    es.eigenvalues()[eig_indices(1)] - es.eigenvalues()[eig_indices(2)], // plate <- voter for curves
                    es.eigenvalues()[eig_indices(2)] // ball <- voter for single points
                    ); 

        tensorVector.push_back(tensorEigenvalues);
        //printf("where: %d \t Stick: %.3f \t Plate: %.3f \t Ball: %.3f\n", 
        //            i, tensorEigenvalues.get<0>(), tensorEigenvalues.get<1>(), tensorEigenvalues.get<2>());
    }
    cout << endl;
    stick=PM::DataPoints::Descriptors::Zero(1, denseField.size());
    plate=PM::DataPoints::Descriptors::Zero(1, denseField.size());
    ball =PM::DataPoints::Descriptors::Zero(1, denseField.size());
    for(unsigned int i=0; i<tensorVector.size(); i++)
    {
        stick(i) = tensorVector[i].get<0>();
        plate(i) = tensorVector[i].get<1>();
        ball(i) = tensorVector[i].get<2>();
    }
    
    // write densegrid to a vector for visualization
    cout	<< "Write to file." << outputname << endl;
    VTKTool::dumpGridCell(outputname, vecDenseCells, cell_size, stick, plate, ball);
    // clean up
    free(h_fieldarraydense);
    free(h_sparsestick);
    free(h_points);

//
    return 1;
}	/* ----------  end of function main  ---------- */
