/*
 * =====================================================================================
 *
 *       Filename:  vtk_voting_example.cpp
 *
 *    Description:  test cuda tensor voting library with vtk input
 *
 *        Version:  1.0
 *        Created:  04/18/2012 07:55 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ming Liu (), ming.liu@mavt.ethz.ch
 *        Company:  ETHZ
 *
 * =====================================================================================
 */

#include <iostream>

//#include "CudaVoting.h"
#include "assert.h"
//#include "vtktool.h"
// libpointmatcher
#include "pointmatcher/PointMatcher.h"
//#include "pointmatcher/Functions.h"
#include "aliases.h"

#include <time.h>
#include <iostream>
#include <fstream>
#define USE_GPU_SPARSE_BALL
#define INT_ENDING 1000000

using namespace PointMatcherSupport;
using namespace std;

//! A dense matrix over ScalarType
typedef Eigen::Matrix<float, Eigen::Dynamic, Eigen::Dynamic> Matrix;
//! A dense integer matrix
typedef Eigen::Matrix<int, Eigen::Dynamic, Eigen::Dynamic> IntMatrix;

/* 
 * ===  FUNCTION  ============================================================
 *         Name:  Main  
 *  Description:  Entry of program
 * ===========================================================================
 */
int main ( int argc, char *argv[] )
{
    DP input;
    if (argc > 2)
    {
        cout << "Input file name: "<< argv[1];
        string filename(argv[1]);
        string output(argv[2]);
        string rate("0.05");
        if (argc > 3)
          rate = argv[3];

        input = PM::loadVTK(filename.c_str());
        cout << "Input size: " << input.features.size()/4 << endl;

        // subsampling here. Because the grid will otherwise too slow.
        PM pm;
        PM::DataPointsFilter* densityFilter;
        densityFilter = pm.DataPointsFilterRegistrar.create("SamplingSurfaceNormalDataPointsFilter",
                    PM::Parameters({
                        {"ratio", "0.98"},
                        {"binSize", "20"},
                        {"keepDensities", "1"}
                        }));

        PM::DataPointsFilter* randSample;
        randSample = pm.DataPointsFilterRegistrar.create("MaxDensityDataPointsFilter",
                    PM::Parameters({
                        {"maxDensity", rate.c_str()}                    
                        }));

        input = densityFilter->filter(input);
        input = randSample->filter(input);
        cout << "Sampled size: " << input.features.size()/4 << endl;

        PM::saveVTK(input, output);
    }
    else
    {
        cout << "usage: ./sample_vtk_points <input.vtk> <output.vtk> <rate>";
    }
    return 1;
}	/* ----------  end of function main  ---------- */
