/*
 * =====================================================================================
 *
 *       Filename:  vtk_voting_example.cpp
 *
 *    Description:  test cuda tensor voting library with vtk input
 *
 *        Version:  1.0
 *        Created:  04/18/2012 07:55 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ming Liu (), ming.liu@mavt.ethz.ch
 *        Company:  ETHZ
 *
 * =====================================================================================
 */

#include <iostream>

//#include "CudaVoting.h"
#include "assert.h"
//#include "vtktool.h"
// libpointmatcher
#include "pointmatcher/PointMatcher.h"
#include "aliases.h"

#include "tensor_voting.h"

#include <time.h>
#include <iostream>
#include <fstream>
using namespace PointMatcherSupport;
using namespace std;
using namespace topomap;

//! A dense matrix over ScalarType
typedef Eigen::Matrix<float, Eigen::Dynamic, Eigen::Dynamic> Matrix;
//! A dense integer matrix
typedef Eigen::Matrix<int, Eigen::Dynamic, Eigen::Dynamic> IntMatrix;

/* 
 * ===  FUNCTION  ============================================================
 *         Name:  Main  
 *  Description:  Entry of program
 * ===========================================================================
 */
int main ( int argc, char *argv[] )
{
    DP input;
    if(argc > 1)
    {
//        cout << "Input file name: "<< argv[1];
//        string filename(argv[1]);
//
//
//        input = PM::loadVTK(filename.c_str());
//        cout << "Input size: " << input.features.size()/4 << endl;
//
//        // subsampling here. Because the grid will otherwise too slow.
//        PM pm;
//        PM::DataPointsFilter* randSample;
//        randSample = pm.DataPointsFilterRegistrar.create("RandomSamplingDataPointsFilter",
//                    PM::Parameters({
//                        {"prob", "0.01"}                    
//                        }));
//        input = randSample->filter(input);
//        cout << "Sampled size: " << input.features.size()/4 << endl;
//
//        PM::saveVTK(input, "sparseFloor.vtk");


        input = PM::loadVTK(argv[1]);
    }
    else
    {
        input = PM::loadVTK("sparseFloor.vtk");
    }

    int numPoints = input.features.size()/4;

    // use CPU to vote for planes
    float sigma = 0.5;
    if (argc > 2)
        sigma = atof(argv[2]);
    TensorVoting voter(sigma); 
    cout << "Set point sets..." << endl;
    voter.setPoints(input);
    cout << "Sparse ball voting (CPU)..." << sigma <<  endl;
    voter.sparse_ball_vote();

    ofstream out("log_sparse_cpu.txt");
    for(int i =0; i<numPoints; i++)
        out << voter.sparseTensors(i) << endl;
    out.close();

    ofstream out1("log_point_cpu.txt");
    out1 << input.features << endl;
    out1.close();

    cout	<< "sparse tensor split..." << endl;
    voter.sparse_tensor_split();

    // save sparse tensor
   PointMatcher<float>::Matrix stick=PM::Matrix::Zero(1, numPoints);
   PointMatcher<float>::Matrix plate=PM::Matrix::Zero(1, numPoints);
   PointMatcher<float>::Matrix ball =PM::Matrix::Zero(1, numPoints);
    for(unsigned int i=0; i<numPoints; i++)
    {
        stick(i) = voter.sparseStick(0,i);
        plate(i) = voter.sparsePlate(0,i*2);
        ball(i) =  voter.sparseBall(i);
    }
    input.addDescriptor("stick", stick);
    input.addDescriptor("plate", plate);
    input.addDescriptor("ball", ball);

    char outname[1024];
    sprintf(outname, "outputSparse_%.2f.vtk", sigma);
    PM::saveVTK(input, outname);

    return 1;

}

